//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//
//
/// \file B2bChamberParameterisation.cc
/// \brief Implementation of the B2bChamberParameterisation class

#include "B2bChamberParameterisation.hh"

#include "G4SystemOfUnits.hh"
#include "G4ThreeVector.hh"
#include "G4Tubs.hh"
#include "G4VPhysicalVolume.hh"

//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo......

B2bChamberParameterisation::B2bChamberParameterisation(G4int noChambers,
                                                       G4double startZ,  //  Z of center of first
                                                       G4double spacingZ,  //  Z spacing of centers
                                                       G4double widthChamber,
                                                       G4double lengthInitial, G4double lengthFinal)
  : G4VPVParameterisation()
{
  fNoChambers = noChambers;
  fStartZ = startZ;
  fHalfWidth = 0.5 * widthChamber;
  fSpacing = spacingZ;
  fRmaxFirst = 0.5 * lengthInitial;
  if (noChambers > 0) {
    fRmaxIncr = 0.5 * (lengthFinal - lengthInitial) / (noChambers - 1);
    if (spacingZ < widthChamber) {
      G4Exception("B2bChamberParameterisation::B2bChamberParameterisation()", "InvalidSetup",
                  FatalException, "Width>Spacing");
    }
  }
}

//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo......

B2bChamberParameterisation::~B2bChamberParameterisation() {}

//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo......

void B2bChamberParameterisation::ComputeTransformation(const G4int copyNo,
                                                       G4VPhysicalVolume* physVol) const
{
  // Note: copyNo will start with zero!
  G4double Zposition = fStartZ + copyNo * fSpacing;
  G4ThreeVector origin(0, 0, Zposition);
  physVol->SetTranslation(origin);
  physVol->SetRotation(0);
}

//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo......

void B2bChamberParameterisation::ComputeDimensions(G4Tubs& trackerChamber, const G4int copyNo,
                                                   const G4VPhysicalVolume*) const
{
  // Note: copyNo will start with zero!
  G4double rmax = fRmaxFirst + copyNo * fRmaxIncr;
  trackerChamber.SetInnerRadius(0);
  trackerChamber.SetOuterRadius(rmax);
  trackerChamber.SetZHalfLength(fHalfWidth);
  trackerChamber.SetStartPhiAngle(0. * deg);
  trackerChamber.SetDeltaPhiAngle(360. * deg);
}

//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo......
