//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//
//
//
// G4PSDoseDeposit
#include "G4PSDoseDeposit.hh"
#include "G4VSolid.hh"
#include "G4VPhysicalVolume.hh"
#include "G4VPVParameterisation.hh"
#include "G4UnitsTable.hh"
#include "G4VScoreHistFiller.hh"

////////////////////////////////////////////////////////////////////////////////
// (Description)
//   This is a primitive scorer class for scoring only energy deposit.
//
//
// Created: 2005-11-14  Tsukasa ASO, Akinori Kimura.
// 2010-07-22   Introduce Unit specification.
// 2020-10-06   Use G4VPrimitivePlotter and fill 1-D histo of dose deposit (x)
//              vs. weight (y) (Makoto Asai)
//
///////////////////////////////////////////////////////////////////////////////

G4PSDoseDeposit::G4PSDoseDeposit(G4String name, G4int depth)
  : G4PSDoseDeposit(name, "Gy", depth) 
{}

G4PSDoseDeposit::G4PSDoseDeposit(G4String name, const G4String& unit,
                                 G4int depth)
  : G4VPrimitivePlotter(name, depth)
  , HCID(-1)
  , EvtMap(nullptr)
{
  SetUnit(unit);
}

G4bool G4PSDoseDeposit::ProcessHits(G4Step* aStep, G4TouchableHistory*)
{
  G4double edep = aStep->GetTotalEnergyDeposit();
  if(edep == 0.)
    return false;

  G4int idx = ((G4TouchableHistory*) (aStep->GetPreStepPoint()->GetTouchable()))
                ->GetReplicaNumber(indexDepth);
  G4double cubicVolume = ComputeVolume(aStep, idx);

  G4double density = aStep->GetTrack()
                       ->GetStep()
                       ->GetPreStepPoint()
                       ->GetMaterial()
                       ->GetDensity();
  G4double dose  = edep / (density * cubicVolume);
  G4double wei   = aStep->GetPreStepPoint()->GetWeight();
  G4int index    = GetIndex(aStep);
  G4double dosew = dose * wei;
  EvtMap->add(index, dosew);

  if(!hitIDMap.empty() && hitIDMap.find(index) != hitIDMap.cend())
  {
    auto filler = G4VScoreHistFiller::Instance();
    if(filler == nullptr)
    {
      G4Exception(
        "G4PSDoseDeposit::ProcessHits", "SCORER0123", JustWarning,
        "G4TScoreHistFiller is not instantiated!! Histogram is not filled.");
    }
    else
    {
      filler->FillH1(hitIDMap[index], dose, wei);
    }
  }

  return true;
}

void G4PSDoseDeposit::Initialize(G4HCofThisEvent* HCE)
{
  EvtMap = new G4THitsMap<G4double>(GetMultiFunctionalDetector()->GetName(),
                                    GetName());
  if(HCID < 0)
  {
    HCID = GetCollectionID(0);
  }
  HCE->AddHitsCollection(HCID, (G4VHitsCollection*) EvtMap);
}

void G4PSDoseDeposit::clear() { EvtMap->clear(); }

void G4PSDoseDeposit::PrintAll()
{
  G4cout << " MultiFunctionalDet  " << detector->GetName() << G4endl;
  G4cout << " PrimitiveScorer " << GetName() << G4endl;
  G4cout << " Number of entries " << EvtMap->entries() << G4endl;
  for(const auto& [copy, dose] : *(EvtMap->GetMap()))
  {
    G4cout << "  copy no.: " << copy
           << "  dose deposit: " << *(dose) / GetUnitValue() << " ["
           << GetUnit() << "]" << G4endl;
  }
}

void G4PSDoseDeposit::SetUnit(const G4String& unit)
{
  CheckAndSetUnit(unit, "Dose");
}

G4double G4PSDoseDeposit::ComputeVolume(G4Step* aStep, G4int idx)
{
  G4VSolid* solid = ComputeSolid(aStep, idx);
  return solid->GetCubicVolume();
}
